#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Meeko
#

from copy import deepcopy
from collections import defaultdict, OrderedDict
import inspect
import json

import numpy as np
from rdkit import Chem
from rdkit.Chem import rdPartialCharges

from .utils import rdkitutils
from .utils import utils

try:
    from openbabel import openbabel as ob
    from .utils import obutils
except ImportError:
    _has_openbabel = False
else:
    _has_openbabel = True

# based on the assumption we are using OpenBabel

# TODO modify so that there are no more dictionaries and only list/arrays (fix me)
# methods like add_x,del_x,get_x will deal with indexing (fix me)
# the goal is to remove dictionaries (fix me)

# (fix me)) provide infrastructure to calculate "differential" modifications,
# like tautomers: the class calculating tautomers will copy and modify the setup
# to swap protons and change the atom types (and possibly rotations?), then
# there will e a function to extract only differences (i.e., atom types and coordinates,
# in the case of tautomers) and store them to make it a switch-state

# TODO use only 1D arrays (C)


# TODO update add_atom to not to specify neighbors, which should be defined only when a bond is created?
# TODO change all attributes_to_copy to have underscore ?

class MoleculeSetup:
    """ mol: molecule structurally prepared with explicit hydrogens

        the setup provides:
            - data storage
            - SMARTS matcher for all atom typers
    """
    # possibly useless here
    attributes_to_copy = [
        "atom_pseudo",
        "atom_ignore",
        "atom_type",
        "atom_true_count",
        "atom_pseudo",
        "element",
        "coord",
        "charge",
        "pdbinfo",
        "chiral",
        "graph",
        "bond",
        "interaction_vector",
        "rings",
        "rings_aromatic",
        "atom_to_ring_id",
        "flexibility_model",
        'history',
        'name',
        ]

    def __init__(self, mol, keep_chorded_rings=False, keep_equivalent_rings=False,
                flexible_amides=False, assign_charges=True, template=None):
        """initialize a molecule template, either from scratch (template is None)
            or by using an existing setup (template is an instance of MoleculeSetup
        """
        stack = inspect.stack()
        the_class = stack[1][0].f_locals["self"].__class__.__name__
        the_method = stack[1][0].f_code.co_name
        #print("I was called by {}.{}()".format(the_class, the_method))
        #print("Setup initialized with:", mol, "template:", template)

        self.mol = mol
        self.atom_pseudo = []
        self.coord = OrderedDict()  # FIXME all OrderedDict shuold be converted to lists?
        self.charge = OrderedDict()
        self.pdbinfo = OrderedDict()
        self.atom_type = OrderedDict()
        self.atom_ignore = OrderedDict()
        self.chiral = OrderedDict()
        self.atom_true_count = 0
        self.graph = OrderedDict()
        self.bond = OrderedDict()
        self.element = OrderedDict()
        self.interaction_vector = OrderedDict()
        self.flexibility_model = {}
        # ring information
        self.rings = {}
        self.rings_aromatic = []
        self.atom_to_ring_id = defaultdict(list)
        self.ring_corners = {}  # used to store corner flexibility
        self.name = None
        # this could be used to keep track of transformations? (corner flipping)
        self.history = []
        if template is None:
            self.process_mol(flexible_amides, assign_charges, keep_chorded_rings, keep_equivalent_rings)
        else:
            if not isinstance(template, MoleculeSetup):
                raise TypeError('FATAL: template must be an instance of MoleculeSetup')
            self.copy_attributes_from(template)

    def process_mol(self, flexible_amides, assign_charges, keep_chorded_rings, keep_equivalent_rings):
        self.atom_true_count = self.get_num_mol_atoms()
        self.name = self.get_mol_name()
        self.init_atom(assign_charges)
        self.perceive_rings(keep_chorded_rings, keep_equivalent_rings)
        self.init_bond(flexible_amides)
        return

    def add_atom(self, idx=None, coord=np.array([0.0, 0.0,0.0], dtype='float'),
            element=None, charge=0.0, atom_type=None, pdbinfo=None, neighbors=None,
            ignore=False, chiral=False, overwrite=False):
        """ function to add all atom information at once;
            every property is optional
        """
        if idx is None:
            idx = len(self.coord)
        if idx in self.coord and not overwrite:
            print("ADD_ATOM> Error: the idx [%d] is already occupied (use 'overwrite' to force)")
            return False
        self.set_coord(idx, coord)
        self.set_charge(idx, charge)
        self.set_element(idx, element)
        self.set_atom_type(idx, atom_type)
        self.set_pdbinfo(idx, pdbinfo)
        if neighbors is None:
            neighbors = []
        self.set_neigh(idx, neighbors)
        self.set_chiral(idx, chiral)
        self.set_ignore(idx, ignore)
        return idx

    def del_atom(self, idx):
        """ remove an atom and update all data associate with it """
        pass
        # coords
        # charge
        # element
        # type
        # neighbor graph
        # chiral
        # ignore
        # update bonds bonds (using the neighbor graph)
        # If pseudo-atom, update other information, too


    # pseudo-atoms
    def add_pseudo(self, coord=np.array([0.0,0.0,0.0], dtype='float'), charge=0.0,
            anchor_list=None, atom_type=None, bond_type=None, rotatable=False,
            pdbinfo=None, directional_vectors=None, ignore=False, chira0=False, overwrite=False):
        """ add a new pseudoatom
            multiple bonds can be specified in "anchor_list" to support the centroids of aromatic rings

            if rotatable, makes the anchor atom rotatable to allow the pseudoatom movement
        """
        idx = self.atom_true_count + len(self.atom_pseudo)
        if idx in self.coord and not overwrite:
            print("ADD_PSEUDO> Error: the idx [%d] is already occupied (use 'overwrite' to force)")
            return False
        self.atom_pseudo.append(idx)
        # add the pseudoatom information to the atoms
        self.add_atom(idx=idx, coord=coord,
                element=0,
                charge=charge,
                atom_type=atom_type,
                pdbinfo=pdbinfo,
                neighbors=[],
                ignore=ignore,
                overwrite=overwrite)
        # anchor atoms
        if not anchor_list is None:
            for anchor in anchor_list:
                self.add_bond(idx, anchor, 0, rotatable, bond_type=bond_type)
        # directional vectors
        if not directional_vectors is None:
            self.add_interaction_vector(idx, directional_vectors)
        return idx

    # Bonds
    def add_bond(self, idx1, idx2, order=0, rotatable=False, in_rings=None, bond_type=None):
        """ bond_type default: 0 (non rotatable) """
        # NOTE: in_ring is used during bond typing to keep endo-cyclic rotatable bonds (e.g., sp3)
        #       as non-rotatable. Possibly, this might be handled by the bond typer itself?
        #       the type would allow it
        # TODO check if in_rings should be checked by this function?
        if in_rings is None:
            in_rings = []
        if not idx2 in self.graph[idx1]:
            self.graph[idx1].append(idx2)
        if not idx1 in self.graph[idx2]:
            self.graph[idx2].append(idx1)
        self.set_bond(idx1, idx2, order, rotatable, in_rings, bond_type)

    # atom types
    def set_atom_type(self, idx, atom_type):
        """ set the atom type for atom index
        atom_type : int or str?
        return: void
        """
        self.atom_type[idx] = atom_type

    def get_atom_type(self, idx):
        """ return the atom type for atom index in the lookup table
        idx : int
        return: str
        """
        return self.atom_type[idx]

    # ignore flag
    def set_ignore(self, idx, state):
        """ set the ignore flag (bool) for the atom
        (used formerged hydrogen)
        idx: int
        state: bool
        """
        self.atom_ignore[idx] = state

    # charge
    def get_charge(self, idx):
        """ return partial charge for atom index
        idx: int

        """
        return self.charge[idx]

    def set_charge(self, idx, charge):
        """ set partial charge"""
        self.charge[idx] = charge

    def get_coord(self, idx):
        """ return coordinates of atom index"""
        return self.coord[idx]

    def set_coord(self, idx, coord):
        """ define coordinates of atom index"""
        self.coord[idx] = coord

    def get_neigh(self, idx):
        """ return atoms connected to atom index"""
        return self.graph[idx]

    def set_neigh(self, idx, neigh_list):
        """ update the molecular graph with the neighbor indices provided """
        if not idx in self.graph:
            self.graph[idx] = []
        for n in neigh_list:
            if not n in self.graph[idx]:
                self.graph[idx].append(n)
            if not n in self.graph:
                self.graph[n] = []
            if not idx in self.graph[n]:
                self.graph[n].append(idx)

    def set_chiral(self, idx, chiral):
        """ set chiral flag for atom """
        self.chiral[idx] = chiral

    def get_chiral(self, idx):
        """ get chiral flag for atom """
        return self.chiral[idx]

    def get_ignore(self, idx):
        """ return if the atom is ignored"""
        return bool(self.atom_ignore[idx])

    def is_aromatic(self, idx):
        """ check if atom is aromatic """
        for r in self.rings_aromatic:
            if idx in r:
                return True
        return False

    def set_element(self, idx, elem_num):
        """ set the atomic number of the atom idx"""
        self.element[idx] = elem_num

    def get_element(self, idx):
        """ return the atomic number of the atom idx"""
        return self.element[idx]

    # def get_atom_ring_count(self, idx):
    #     """ return the number of rings to which this atom belongs"""
    #     # FIXME this should be replaced by self.get_atom_rings()
    #     return len(self.atom_to_ring_id[idx])

    def get_atom_rings(self, idx):
        # FIXME this should replace self.get_atom_ring_count()
        """ return the list of rings to which the atom idx belongs"""
        if idx in self.atom_to_ring_id:
            return self.atom_to_ring_id[idx]
        return []

    def get_atom_indices(self, true_atoms_only=False):
        """ return the indices of the atoms registered in the setup
            if 'true_atoms_only' are requested, then pseudoatoms are ignored
        """
        indices = list(self.coord.keys())
        if true_atoms_only:
            return [ x for x in indices if not x in self.atom_pseudo ]
        return indices

    # interaction vectors
    def add_interaction_vector(self, idx, vector_list):
        """ add vector list to list of directional interaction vectors for atom idx"""
        if not idx in self.interaction_vector:
            self.interaction_vector[idx] = []
        for vec in vector_list:
            self.interaction_vector[idx].append(vec)

    # TODO evaluate if useful
    def _get_attrib(self, idx, attrib, default=None):
        """ generic function to provide a default for retrieving properties and returning standard values """
        return getattr(self, attrib).get(idx, default)

    def get_interaction_vector(self, idx):
        """ get list of directional interaction vectors for atom idx"""
        return self.interaction_vector[idx]

    def del_interaction_vector(self, idx):
        """ delete list of directional interaction vectors for atom idx"""
        del self.interaction_vector[idx]

    def set_pdbinfo(self, idx, data):
        """ add PDB data (resname/num, atom name, etc.) to the atom """
        self.pdbinfo[idx] = data

    def get_pdbinfo(self, idx):
        """ retrieve PDB data (resname/num, atom name, etc.) to the atom """
        return self.pdbinfo[idx]

    def set_bond(self, idx1, idx2, order=None, rotatable=None, in_rings=None, bond_type=None):
        """ populate bond lookup table with properties
            bonds are identified by any tuple of atom indices
            the function generates the canonical bond id

            order      : int
            rotatable  : bool
            in_rings   : list (rings to which the bond belongs)
            bond_type  : int
        """
        bond_id = self.get_bond_id(idx1, idx2)
        if order is None:
            order = 0
        if rotatable is None:
            rotatable = False
        if in_rings is None:
            in_rings = []
        self.bond[bond_id] = {'bond_order': order,
                              'type': bond_type,
                              'rotatable': rotatable,
                              'in_rings': in_rings}

    def del_bond(self, idx1, idx2):
        """ remove a bond from the lookup table """
        bond_id = self.get_bond_id(idx1, idx2)
        del self.bond[bond_id]
        self.graph[idx1].remove(idx2)
        # TODO check if we want to delete nodes that have no connections (we might want to keep them)
        if not self.graph[idx1]:
            del self.graph[idx1]
        self.graph[idx2].remove(idx1)
        if not self.graph[idx2]:
            del self.graph[idx2]

    def get_bond(self, idx1, idx2):
        """ return properties of a bond in the lookup table
            if the bond does not exist, None is returned

            idx1, idx2 : int

            return: dict or voidko
        """
        bond_idx = self.get_bond_id(idx1, idx2)
        try:
            return self.bond[bond_idx]
        except IndexError:
            return None

    @staticmethod
    def get_bond_id(idx1, idx2):
        """ used to generate canonical bond id from a pair of nodes in the graph"""
        idx_min = min(idx1, idx2)
        idx_max = max(idx1, idx2)
        return (idx_min, idx_max)

    # replaced by
    # def ring_atom_to_ring(self, arg):
    #     return self.atom_to_ring_id[arg]

    def get_bonds_in_ring(self, ring):
        """ input: 'ring' (list of atom indices)
            returns list of bonds in ring, each bond is a pair of atom indices
        """
        n = len(ring)
        bonds = []
        for i in range(n):
            bond = (ring[i], ring[(i+1) % n])
            bond = (min(bond), max(bond))
            bonds.append(bond)
        return bonds

    def walk_recursive(self, idx, collected=None, exclude=None):
        """ walk molecular graph and return subgraphs that are bond-connected"""
        if collected is None:
            collected = []
        if exclude is None:
            exclude = []
        for neigh in self.get_neigh(idx):
            if neigh in exclude:
                continue
            collected.append(neigh)
            exclude.append(neigh)
            self.walk_recursive(neigh, collected, exclude)
        return collected

    def copy_attributes_from(self, template):
        """ copy attributes to duplicate the template setup
        NOTE: the molecule will always keep the original setup (i.e., template)"""
        # TODO enable some kind of plugin system here too, to allow other objects to
        # add attributes?
        # TODO although, this would make the setup more fragile-> better have attributes
        # explicitely added here, and that's it
        for attr in self.attributes_to_copy:
            attr_copy = deepcopy(getattr(template, attr))
            setattr(self, attr, attr_copy)
        # TODO possible BUG? the molecule is shared by the different setups
        #      if one of them alters the molecule, properties will not be the same
        self.mol = template.mol

    def merge_hydrogen(self):
        """ standard method to merge hydrogens bound to carbons"""
        merged = 0

        for a, neigh_list in list(self.graph.items()):
            # look for hydrogens
            if not self.get_element(a) == 1:
                continue

            hydrogen_charge = self.get_charge(a)

            for n in neigh_list:
                # look for carbons
                if self.get_element(n) == 6:
                    merged += 1
                    carbon_charge = self.get_charge(n)
                    # carbon adsorbs the final charge
                    self.set_charge(n, carbon_charge + hydrogen_charge)
                    # flag hydrogen to be ignored and set its charge to 0
                    self.set_charge(a, 0)
                    self.set_ignore(a, True)

    def copy(self):
        """ return a copy of the current setup"""
        raise NotImplementedError("This method must be overloaded by inheriting class")

    def init_atom(self):
        """ iterate through molecule atoms and build the atoms table """
        raise NotImplementedError("This method must be overloaded by inheriting class")

    def perceive_rings(self, keep_chorded_rings, keep_equivalent_rings):
        """ populate the rings and aromatic rings tableshe atoms table:
        self.rings_aromatics : list
            contains the list of ring_id items that are aromatic
        self.rings: dict
            store information about rings, like if they have corners that can
            be flipped and the graph of atoms that belong to them:

                self.rings[ring_id] = {
                                'corner_flip': False
                                'graph': {}
                                }

            The atom is built using the `walk_recursive` method
        self.ring_atom_to_ring_id: dict
            mapping of each atom belonginig to the ring: atom_idx -> ring_id
        """

        def isRingAromatic(ring_atom_indices):
            for atom_idx1, atom_idx2 in self.get_bonds_in_ring(ring_atom_indices):
                bond = self.mol.GetBondBetweenAtoms(atom_idx1, atom_idx2)
                if not bond.GetIsAromatic():
                    return False
            return True

        hjk_ring_detection = utils.HJKRingDetection(self.graph) 
        rings = hjk_ring_detection.scan(keep_chorded_rings, keep_equivalent_rings) # list of tuples of atom indices
        for ring_atom_idxs in rings:
            if isRingAromatic(ring_atom_idxs):
                self.rings_aromatic.append(ring_atom_idxs)
            self.rings[ring_atom_idxs] = {'corner_flip':False}
            graph = {}
            for atom_idx in ring_atom_idxs:
                self.atom_to_ring_id[atom_idx].append(ring_atom_idxs)
                # graph of atoms affected by potential ring movements
                graph[atom_idx] = self.walk_recursive(atom_idx, collected=[], exclude=list(ring_atom_idxs))
            self.rings[ring_atom_idxs]['graph'] = graph

    def init_bond(self, flexible_amides):
        """ iterate through molecule bonds and build the bond table (id, table)
            INPUT
                flexible_amides: bool
                    this flag is used to define if primary and secondary amides
                    need to be considered flexible or not (bond_order= 999)
            CALCULATE
                bond_order: int
                    0       : pseudo-bond (for pseudoatoms)
                    1,2,3   : single-triple bond
                    5       : aromatic
                    999     : rigid

                if bond is in ring (both start and end atom indices in the bond are in the same ring)

            SETUP OPERATION
                Setup.add_bond(idx1, idx2, order, in_rings=[])
        """
        raise NotImplementedError("This method must be overloaded by inheriting class")

    def get_mol_name(self):
        raise NotImplementedError("This method must be overloaded by inheriting class")

    def find_pattern(self, smarts):
        raise NotImplementedError("This method must be overloaded by inheriting class")

    def get_smiles_and_order(self):
        raise NotImplementedError("This method must be overloaded by inheriting class")


class RDKitMoleculeSetup(MoleculeSetup):

    def get_smiles_and_order(self):

        # 3D SDF files written by other toolkits (OEChem, ChemAxon)
        # seem to not include the chiral flag in the bonds block, only in
        # the atoms block. RDKit ignores the atoms chiral flag as per the
        # spec. When reading SDF (e.g. from PubChem/PDB),
        # we may need to have RDKit assign stereo from coordinates, see:
        # https://sourceforge.net/p/rdkit/mailman/message/34399371/
        mol_noH = Chem.RemoveHs(self.mol) # imines (=NH) may become chiral
        atomic_num_mol_noH = [atom.GetAtomicNum() for atom in mol_noH.GetAtoms()]
        noH_to_H = []
        num_H_in_noH = 0 # e.g. stereo imines [H]/N=C keep [H] after RemoveHs()
        for (index, atom) in enumerate(self.mol.GetAtoms()):
            if atom.GetAtomicNum() == 1: continue
            for i in range(len(noH_to_H), len(atomic_num_mol_noH)):
                if atomic_num_mol_noH[i] > 1: break
                noH_to_H.append('H')
            noH_to_H.append(index)
        extra_hydrogens = len(atomic_num_mol_noH) - len(noH_to_H)
        if extra_hydrogens > 0:
            assert(set(atomic_num_mol_noH[len(noH_to_H):]) == {1})
            noH_to_H.extend(['H'] * extra_hydrogens)

        # map indices of explicit hydrogens, e.g. stereo imine [H]/N=C
        for index in range(len(noH_to_H)):
            if noH_to_H[index] != 'H': continue
            h_atom = mol_noH.GetAtomWithIdx(index)
            assert(h_atom.GetAtomicNum() == 1)
            parents = h_atom.GetNeighbors()
            assert(len(parents) == 1)
            num_h_in_parent = len([a for a in parents[0].GetNeighbors() if a.GetAtomicNum() == 1])
            if num_h_in_parent != 1:
                msg = "Can't handle %d explicit H for each heavy atomin noH mol.\n" % num_h_in_parent
                msg += "Was expecting only imines [H]N=\n"
                raise RuntimeError(msg)
            parent_index_in_mol_with_H = noH_to_H[parents[0].GetIdx()]
            parent_in_mol_with_H = self.mol.GetAtomWithIdx(parent_index_in_mol_with_H)
            h_in_mol_with_H = [a for a in parent_in_mol_with_H.GetNeighbors() if a.GetAtomicNum() == 1]
            if len(h_in_mol_with_H) != 1:
                msg = "Can't handle %d explicit H for each heavy atomin noH mol.\n" % len(h_in_mol_with_H)
                msg += "Was expecting only imines [H]N=\n"
                raise RuntimeError(msg)
            noH_to_H[index] = h_in_mol_with_H[0].GetIdx()

        smiles = Chem.MolToSmiles(mol_noH)
        order_string = mol_noH.GetProp("_smilesAtomOutputOrder")
        order_string = order_string.replace(',]', ']') # remove trailing comma
        order = json.loads(order_string) # mol_noH to smiles
        order = list(np.argsort(order))
        order = {noH_to_H[i]: order[i]+1 for i in range(len(order))} # 1-index
        return smiles, order

    def find_pattern(self, smarts):
        p = Chem.MolFromSmarts(smarts)
        return self.mol.GetSubstructMatches(p)

    def get_mol_name(self):
        if self.mol.HasProp("_Name"):
            return self.mol.GetProp("_Name")
        else:
            return None

    def get_num_mol_atoms(self):
        return self.mol.GetNumAtoms()

    def get_equivalent_atoms(self):
       return list(Chem.CanonicalRankAtoms(self.mol, breakTies=False))

    def init_atom(self, assign_charges):
        """ initialize the atom table information """
        # extract the coordinates
        c = self.mol.GetConformers()[0]
        coords = c.GetPositions()
        # extract/generate charges
        if assign_charges:
            rdPartialCharges.ComputeGasteigerCharges(self.mol)
            charges = [a.GetDoubleProp('_GasteigerCharge') for a in self.mol.GetAtoms()]
        else:
            charges = [0.0] * self.mol.GetNumAtoms()
        # perceive chirality
        # TODO check consistency for chiral model between OB and RDKit
        chiral_info = {}
        for data in Chem.FindMolChiralCenters(self.mol, includeUnassigned=True):
            chiral_info[data[0]] = data[1]
        # register atom
        for a in self.mol.GetAtoms():
            idx = a.GetIdx()
            chiral = False
            if idx in chiral_info:
                chiral = chiral_info[idx]
            self.add_atom(idx,
                    coord=coords[idx],
                    element=a.GetAtomicNum(),
                    charge=charges[idx],
                    atom_type=None,
                    pdbinfo = rdkitutils.getPdbInfoNoNull(a),
                    neighbors = [n.GetIdx() for n in a.GetNeighbors()],
                    chiral=False,
                    ignore=False)

    def init_bond(self, flexible_amides):
        """ initialize bond information """
        amide_smarts = "[$([NX3][#1])][$([CX3](=[OX1])[#6])]" # does not match tertiary amides
        amide_bonds = [set(f) for f in self.find_pattern(amide_smarts)]
        for b in self.mol.GetBonds():
            idx1 = b.GetBeginAtomIdx()
            idx2 = b.GetEndAtomIdx()
            bond_order = int(b.GetBondType())
            # fix the RDKit aromatic type (FIXME)
            if bond_order == 12: # aromatic
                bond_order = 5
            if bond_order == 1:
                rotatable = True
                if flexible_amides:
                    continue
                # check if bond is a tertiary amide bond
                if set((idx1, idx2)) in amide_bonds:
                    # TODO check what is the best to keep here (rotatable or bond_order)
                    rotatable=False
                    bond_order = 999
            else:
                rotatable = False
            idx1_rings = set(self.get_atom_rings(idx1))
            idx2_rings = set(self.get_atom_rings(idx2))
            in_rings = list(set.intersection(idx1_rings, idx2_rings))
            self.add_bond(idx1, idx2, order=bond_order, rotatable=rotatable, in_rings=in_rings)

    def copy(self):
        """ return a copy of the current setup"""
        return RDKitMoleculeSetup(self.mol, template=self)


class OBMoleculeSetup(MoleculeSetup):

    def get_mol_name(self):
        return self.mol.GetTitle()

    def find_pattern(self, smarts):
        obsmarts = ob.OBSmartsPattern()
        obsmarts.Init(smarts)
        found = obsmarts.Match(self.mol)
        output = []
        if found:
            for x in obsmarts.GetUMapList():
                output.append([y-1 for y in x])
        return output

    def get_num_mol_atoms(self):
        return self.mol.NumAtoms()

    def get_equivalent_atoms(self):
        raise NotImplementedError 
    
    def init_atom(self, assign_charges):
        """initialize atom data table"""
        for a in ob.OBMolAtomIter(self.mol):
            # TODO fix this to be zero-based?
            partial_charge = a.GetPartialCharge() * float(assign_charges)
            self.add_atom(a.GetIdx() - 1,
                    coord=np.asarray(obutils.getAtomCoords(a), dtype='float'),
                    element=a.GetAtomicNum(),
                    charge=partial_charge,
                    atom_type=None,
                    pdbinfo = obutils.getPdbInfoNoNull(a),
                    neighbors=[x.GetIdx() - 1 for x in ob.OBAtomAtomIter(a)],
                    ignore=False, chiral=a.IsChiral())
            # TODO check consistency for chiral model between OB and RDKit

    def init_bond(self, flexible_amides):
        """initialize bond data table"""
        for b in ob.OBMolBondIter(self.mol):
            idx1 = b.GetBeginAtomIdx() - 1
            idx2 = b.GetEndAtomIdx() - 1
            bond_order = b.GetBondOrder()
            if b.IsAromatic():
                bond_order = 5
            if b.IsAmide() and not b.IsTertiaryAmide():
                if not flexible_amides:
                    bond_order = 999
            # check if bond is a ring bond, i.e., both atoms belongs to the same ring
            idx1_rings = set(self.get_atom_rings(idx1))
            idx2_rings = set(self.get_atom_rings(idx2))
            in_rings = list(set.intersection(idx1_rings, idx2_rings))
            self.add_bond(idx1, idx2, order=bond_order, in_rings=in_rings)

    def copy(self):
        """ return a copy of the current setup"""
        return OBMoleculeSetup(self.mol, template=self)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Meeko macrocycle builder
#

import os
import sys
from collections import defaultdict
from operator import itemgetter


class FlexMacrocycle:
    def __init__(self, min_ring_size=7, max_ring_size=33, double_bond_penalty=50, max_breaks=4):
        """Initialize macrocycle typer.

        Args:
            min_ring_size (int): minimum size of the ring (default: 7)
            max_ring_size (int): maximum size of the ring (default: 33)
            double_bond_penalty (float)

        """
        self._min_ring_size = min_ring_size
        self._max_ring_size = max_ring_size
        # accept also double bonds (if nothing better is found)
        self._double_bond_penalty = double_bond_penalty
        self.max_breaks = max_breaks

        self.setup = None
        self.breakable_rings = None
        self._conj_bond_list = None
        self.breakable_bonds = None

    def collect_rings(self, setup):
        """ get non-aromatic rings of desired size and
            list bonds that are part of unbreakable rings

            Bonds belonging to rigid cycles can't be deleted or
            made rotatable even if they are part of a breakable ring
        """
        breakable_rings = []
        rigid_rings = []
        for ring_id in list(setup.rings.keys()): # ring_id are the atom indices in each ring
            size = len(ring_id)
            if ring_id in setup.rings_aromatic:
                rigid_rings.append(ring_id)
            elif size < self._min_ring_size:
                rigid_rings.append(ring_id)
                # do not add rings > _max_ring_size to rigid_rings
                # because bonds in rigid rings will not be breakable 
                # and these bonds may also belong to breakable rings
            elif size <= self._max_ring_size:
                breakable_rings.append(ring_id)

        bonds_in_rigid_cycles = set()
        for ring_atom_indices in rigid_rings:
            for bond in setup.get_bonds_in_ring(ring_atom_indices):
                bonds_in_rigid_cycles.add(bond)

        return breakable_rings, bonds_in_rigid_cycles 

    def _detect_conj_bonds(self):
        """ detect bonds in conjugated systems
        """
        # TODO this should be removed once atom typing will be done
        conj_bond_list = []
        # pattern = "[R0]=[R0]-[R0]=[R0]" # Does not match conjugated bonds inside  the macrocycle?
        pattern = '*=*[*]=,#,:[*]' # from SMARTS_InteLigand.txt
        found = self.setup.find_pattern(pattern)
        for f in found:
            bond = (f[1], f[2])
            bond = (min(bond), max(bond))
            conj_bond_list.append(bond)
        return conj_bond_list

    def _score_bond(self, bond):
        """ provide a score for the likeness of the bond to be broken"""
        atom_idx1 = bond[0]
        atom_idx2 = bond[1]
        score = 100

        bond_order = self.setup.bond[bond]['bond_order']
        if bond_order not in [1, 2, 3]: # aromatic, double, made rigid explicitly (order=1.1 from --rigidify)
            return -1
        if not self.setup.get_element(atom_idx1) == 6 or self.setup.is_aromatic(atom_idx1):
            return -1
        if not self.setup.get_element(atom_idx2) == 6 or self.setup.is_aromatic(atom_idx2):
            return -1
        # triple bond tolerated but not preferred (TODO true?)
        if bond_order == 3:
            score -= 30
        elif (bond_order == 2):
            score -= self._double_bond_penalty
        if bond in self._conj_bond_list:
            score -= 30
        # discourage chiral atoms
        if self.setup.get_chiral(atom_idx1) or self.setup.get_chiral(atom_idx2):
            score -= 20
        return score

    def get_breakable_bonds(self, bonds_in_rigid_rings):
        """ find breaking points for rings
            following guidelines defined in [1]
            The optimal bond has the following properties:
            - does not involve a chiral atom
            - is not double/triple (?)
            - is between two carbons
            (preferably? we can now generate pseudoAtoms on the fly!)
            - is a bond present only in one ring

             [1] Forli, Botta, J. Chem. Inf. Model., 2007, 47 (4)
              DOI: 10.1021/ci700036j
        """
        breakable = {}
        for ring_atom_indices in self.breakable_rings:
            for bond in self.setup.get_bonds_in_ring(ring_atom_indices):
                score = self._score_bond(bond)
                if score > 0 and bond not in bonds_in_rigid_rings:
                    breakable[bond] = {'score': score}
        return breakable

    def search_macrocycle(self, setup):
        """Search for macrocycle in the molecule

        Args:
            setup : MoleculeSetup object

        """
        self.setup = setup

        self.breakable_rings, bonds_in_rigid_rings = self.collect_rings(setup)
        self._conj_bond_list = self._detect_conj_bonds()
        self.breakable_bonds = self.get_breakable_bonds(bonds_in_rigid_rings)
        break_combo_data = self.combinatorial_break_search(self.breakable_bonds)
        return break_combo_data, bonds_in_rigid_rings

    def combinatorial_break_search(self, breakable_bonds):
        """ enumerate all combinations of broken bonds
            once a bond is broken, it will break one or more rings
            subsequent bonds will be pulled from intact (unbroken) rings
            
            the number of broken bonds may be variable
            returns only combinations of broken bonds that break the max number of broken bonds
        """

        max_breaks = self.max_breaks
        break_combos = self._recursive_break(self.breakable_rings, max_breaks, breakable_bonds, set(), [])
        break_combos = list(break_combos) # convert from set
        max_broken_bonds = 0
        output_break_combos = [] # found new max, discard prior data
        output_bond_scores = []
        output_broken_rings = []
        for broken_bonds in break_combos:
            n_broken_bonds = len(broken_bonds)
            bond_score = sum([breakable_bonds[bond]['score'] for bond in broken_bonds])
            broken_rings = self.get_broken_rings(self.breakable_rings, broken_bonds)
            if n_broken_bonds > max_broken_bonds:
                max_broken_bonds = n_broken_bonds
                output_break_combos = [] # found new max, discard prior data
                output_bond_scores = []
                output_broken_rings = []
            if n_broken_bonds == max_broken_bonds:
                output_break_combos.append(broken_bonds)
                output_bond_scores.append(bond_score)
                output_broken_rings.append(broken_rings)
        break_combo_data = {"bond_break_combos": output_break_combos,
                            "bond_break_scores": output_bond_scores,
                            "broken_rings": output_broken_rings}
        return break_combo_data


    def _recursive_break(self, rings, max_breaks, breakable_bonds, output=set(), broken_bonds=[]):
        if max_breaks == 0:
            return output
        unbroken_rings = self.get_unbroken_rings(rings, broken_bonds)
        atoms_in_broken_bonds = atoms_in_broken_bonds = [atom_idx for bond in broken_bonds for atom_idx in bond]
        for bond in breakable_bonds:
            if bond[0] in atoms_in_broken_bonds or bond[1] in atoms_in_broken_bonds:
                continue # each atom can be in only one broken bond
            is_bond_in_ring = False
            for ring in unbroken_rings:
                if bond in self.setup.get_bonds_in_ring(ring):
                    is_bond_in_ring = True
                    break
            if is_bond_in_ring:
                current_broken_bonds = [(a, b) for (a, b) in broken_bonds + [bond]]
                num_unbroken_rings = len(self.get_unbroken_rings(rings, current_broken_bonds))
                data_row = tuple(sorted([(a, b) for (a, b) in current_broken_bonds]))
                output.add(data_row)
                if num_unbroken_rings > 0:
                    output = self._recursive_break(rings, max_breaks-1, breakable_bonds,
                                                   output, current_broken_bonds)
        return output

    
    def get_unbroken_rings(self, rings, broken_bonds):
        unbroken = []
        for ring in rings:
            is_unbroken = True
            for bond in broken_bonds:
                if bond in self.setup.get_bonds_in_ring(ring): # consider precalculating bonds
                    is_unbroken = False
                    break # pun intended
            if is_unbroken:
                unbroken.append(ring)        
        return unbroken

    def get_broken_rings(self, rings, broken_bonds):
        broken_rings = []
        for ring in rings:
            is_broken = False
            for bond in broken_bonds:
                if bond in self.setup.get_bonds_in_ring(ring): # consider precalculating bonds
                    is_broken = True
                    break # pun intended
            if is_broken:
                broken_rings.append(ring)        
        return broken_rings


    def show_macrocycle_scores(self, setup):
        print("Warning: not showing macrocycle scores, check implementation.")
        return
        if setup is not None:
            print("\n==============[ MACROCYCLE SCORES ]================")
            bond_by_ring = defaultdict(list)

            for bond_id, data in list(setup.ring_bond_breakable.items()):
                ring_id = data['ring_id']
                bond_by_ring[ring_id].append(bond_id)

            for ring_id, bonds in list(bond_by_ring.items()):
                data = []
                print("-----------[ ring id: %s | size: %2d ]-----------" % (",".join([str(x) for x in ring_id]), len(ring_id)))

                for b in bonds:
                    score = setup.ring_bond_breakable[b]['score']
                    data.append((b, score))

                data = sorted(data, key=itemgetter(1), reverse=True)

                for b_count, b in enumerate(data):
                    begin = b[0][0]
                    end = b[0][1]
                    # bond = self.mol.setup.get_bond(b[0][0], b[0][1])
                    # begin = bond.GetBeginAtomIdx()
                    # end = bond.GetEndAtomIdx()
                    info = (b_count, begin, end, b[1], "#" * int(b[1] / 5), "-" * int(20 - b[1] / 5))
                    print("[ %2d] Bond [%3d --%3d] s:%3d [%s%s]" % info)




from collections import namedtuple
from rdkit import Chem
from rdkit.Chem import AllChem, rdGeometry
import prody
import warnings


from collections import defaultdict
def nested_dict():
    return defaultdict(nested_dict)



# class CovLigandPrepared(object):

#     """ container class to store the prepared ligand molecule with associated information"""
#     def __init__(self, mol, res_id, at_names, smarts, indices, smarts_indices):
#         self.mol = mol
#         self.res_id = res_id
#         self.at_names = at_names
#         self.smarts = smarts
#         self.indices = indices
#         self.smarts_indices = smarts_indices

#     @property
#     def label(self):
#         """ DOCUMENTATION GOES HERE """
#         pass

#     @property
#     def label(self):
#         """ DOCUMENTATION GOES HERE """
#         pass





CovLigandPrepared = namedtuple("CovalentLigandPrepared", ["mol", "res_id", "at_names", "smarts", "smarts_indices", "indices", "label"])


class CovalentBuilder(object):
    """ Class to perform structural alignments of ligands containing the target side chain attached
        to run tethered covalent dockings

        The class is instantiated for a given target, with a list of one or more residues, then
        ligands can be processed sequentially ( CovalentBuilder.process() )

            receptor_mol    :       ProDy molecule
            residue_string  :       string specifying residues to process. Examples:
                                    ":LYS:"              (all LYS, default atom names: CA,CB)
                                    "A:LYS:204:CA,CB"
                                    "A:HIS::ND1,CE1"     (all HIS on chain A)

        (ProDy regular expressions could be also injected here).
    """
    def __init__(self, receptor_mol, residue_string):
        self.rec = receptor_mol
        selection_tuple = self.parse_residue_string(residue_string, force_CA_CB=True)
        # selection tuple: (chain, res, num, atname1, atname2)
        # only 'res' and 'atname[1|2]' are required, the rest is optional,
        # e.g: (None, res, None, 'CA', 'CB')
        out = self._generate_prody_selection(selection_tuple)
        self._compact_selection(out)

    def _generate_prody_selection(self, residue):
        """ generate the string to perform a Prody.Selection """
        # sel_string = 'chid %s AND resname %s AND resnum %s AND (name %s or name %s)'
        sel_string = []
        # out = []
        chid, res_type, res_num, atname1, atname2 = residue
        if not chid == "":
            sel_string.append("chid %s" % chid)
        sel_string.append('resname %s' % res_type)
        if not res_num == "":
            sel_string.append("resnum %s" % res_num)
        sel_string.append("(name %s or name %s)" % (atname1, atname2))
        sel_string = " and ".join(sel_string)
        #print("CovalentBuilder> searching for residue:",sel_string)
        found = self.rec.select( sel_string )
        if found is None:
            print("ERROR: no residue found with the following specification: chain[%s] residue[%s] number[%s] atom names [%s,%s]"% (
                chid, res_type, res_num, atname1, atname2))
            raise ValueError
        return (found, atname1, atname2)
        # returnsout

    def _compact_selection(self, sel_info, allow_missing=False):
        """ process a ProDy selection and return a dictionary with the structure { res_id : (at1_coord, at2_coord), ...} """
        self.residues = {}
        # for (sel, at1, at2) in sel_info:
        sel, at1, at2 = sel_info
        pairs = (at1, at2)
        chains = sel.getChids()
        res = sel.getResnames()
        num = sel.getResnums()
        names = sel.getNames()
        coords = sel.getCoords()
        for i in range(len(chains)):
            res_id = (chains[i], res[i], num[i])
            if not res_id in self.residues:
                self.residues[res_id] = [None, None]
            # preserve the order of the coordinates
            # as specified by the atom names order
            idx = pairs.index(names[i])
            self.residues[res_id][idx] = {'coords':coords[i], 'atname':names[i]}
        res_id_list = list(self.residues.keys())
        for res_id in res_id_list:
            if None in self.residues[res_id]:
                c,r,n = res_id
                f = [ x['atname'] if not x is None else "None" for x in  self.residues[res_id]]
                print("WARNING: one or more atoms are missing in residue %s:%s%d (requested: %s,%s | found: %s,%s)" % (c,r,n,at1,at2, f[0], f[1]) )
                if not allow_missing:
                    raise ValueError
                del self.residues[res_id]
        # from pprint import pprint as pp
        # pp(self.residues)

    def process(self, ligand, smarts=None, smarts_indices=None, indices=None, first_only=False):
        """ process the ligand for the residue(s) specified for the current receptor"""
        if (smarts is None) and (indices is None):
            raise ValueError("Specify at least one criterion, either SMARTS pattern or atom indices (2)")
        # if SMARTS are specified, use that to define (or override) indices
        if not smarts is None:
            indices = self.find_smarts(ligand, smarts, smarts_indices, first_only)
        if len(indices) == 0:
            warnings.warn("SMARTS pattern didn't match any atoms", RuntimeWarning)
        #print("CovalentBuilder> Generating %d ouput alignments (%d residues)" % (len(indices)*len(self.residues), len(self.residues) ))
        # perform alignments
        for i, idx_pair in enumerate(indices):
            for res_info, res_coord in self.residues.items():
                if len(indices)>1:
                    counter = "_%d" % (i+1)
                else:
                    counter = ""
                chain, res, num, = res_info
                at_names = [ x['atname'] for x in res_coord ]
                coord = [ x['coords'] for x in res_coord ]
                label = "%s:%s%s%s" % (chain, res, num, counter)
                mol = self.transform(ligand, idx_pair, coord)
                yield CovLigandPrepared(mol, res_info, at_names, smarts, smarts_indices, idx_pair,label)

    def find_smarts(self, mol, smarts, smarts_indices, first_only):
        """ find occurrences of the SMARTS indices atoms in the requested SMARTS"""
        indices = []
        patt = Chem.MolFromSmarts(smarts)
        smarts_size = patt.GetNumAtoms()
        if smarts_indices[0] >= smarts_size or smarts_indices[1] >= smarts_size:
            raise ValueError("SMARTS index exceeds number of atoms in SMARTS (%d)" % (smarts_size))
        found = mol.GetSubstructMatches(patt)
        #print("CovalentBuilder> ligand patterns found: ", found, "[ use only first: %s ]" % first_only)
        if len(found)>1 and first_only:
            print("WARNING: the specified ligand pattern returned more than one match: [%d] (potential ambiguity?)" % len(found))
        for f in found:
            #print("CovalentBuilder> processing:", f, "with ", smarts_indices)
            indices.append([f[x] for x in smarts_indices])
            #print("CovalentBuilder> ligand indices stored:", indices)
            if first_only:
                return indices
        return indices

    def transform(self, ligand, index_pair, coord):
        """ generate translatead and aligned molecules for each of the indices requested
            and for all the residues defined in the class constructor

            SOURCE: https://sourceforge.net/p/rdkit/mailman/message/36750909/
        """
        # make a copy of the ligand
        # TODO: maybe define new conformers?
        mol = Chem.Mol(ligand)
        target = Chem.MolFromSmiles("CC")
        # add hydrogens
        target = Chem.AddHs(target)
        # generate 3D coords
        AllChem.EmbedMolecule(target)
        # get the first conformer
        conf = target.GetConformer()
        # set coordinates to the ones of the actual target residue
        c1 = rdGeometry.Point3D( coord[0][0], coord[0][1], coord[0][2] )
        c2 = rdGeometry.Point3D( coord[1][0], coord[1][1], coord[1][2] )
        conf.SetAtomPosition(0, c1)
        conf.SetAtomPosition(1, c2)
        # perform alignment
        Chem.rdMolAlign.AlignMol(mol, target, -1, -1,[(index_pair[0],0), (index_pair[1], 1)] )
        return mol

    @classmethod
    def parse_residue_string(cls, string, force_CA_CB=True):
        chain, res, num, *atom_names = string.split(":")
        #print("PARSED c:%s r:%s n:%s AAA:%s" % (chain, res, num, atom_names))
        if not num == "":
            if not num.isdigit():
                raise ValueError('NUM must be an integer in "CHAIN:RES:NUM" or "CHAIN:RES:NUM:ATOM1,ATOM2"')
        # parse atom names
        #  TODO clean all specifications
        if atom_names == []:
            atom_names = "CA,CB"
        else:
            atom_names = atom_names[0]
        # trigger error if comma not found
        a1, a2 = atom_names.split(",")
        # check that the legacy_format flag is not used
        if force_CA_CB and (not a1 == 'CA' or not a2 == 'CB'):
            msg  = "Atom names expected to be CA,CB but got %s,%s instead.\n" % (a1, a2)
            raise RuntimeError(msg)
        residue = (chain, res, num, a1, a2)
        return residue



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Raccoon
#

# TODO this should be in its own obabelutils module
# from collections import namedtuple

import numpy as np
from openbabel import openbabel as ob

from . import geomutils
from . import utils
from . import pdbutils
from .utils import mini_periodic_table

# named tuple to contain information about an atom
# PDBAtomInfo = namedtuple('PDBAtomInfo', "name resName resNum chain")
# PDBResInfo  = namedtuple('PDBResInfo',       "resName resNum chain")


def getAtomIdxCoords(obmol, atom_idx):
    """return coordinates of atom idx """
    atom = obmol.GetAtom(atom_idx)
    return getAtomCoords(atom)


def getAtomCoords(atom):
    """ convert an OB atom into a numpy vector of coordinates """
    return np.array([atom.GetX(), atom.GetY(), atom.GetZ()], 'f')


def getCoordsFromAtomIndices(obmol, atomIdxList):
    """ extract coordinates for requested atom indices (return Numpy.array)"""
    coord = []
    for idx in atomIdxList:
        a = obmol.GetAtom(idx)
        coord.append(getAtomCoords(a))
    return np.array(coord)

def getAtoms(obmol):
    return ob.OBMolAtomIter(obmol)

def getAtomRes(atom):
    """ retrieve residue info about the atom """
    r = atom.GetResidue()
    data = {'num': r.GetNum(), 'name': r.GetName(), 'chain': r.GetChain()}
    return data


def atomsCentroid(obmol, atomIndices):
    """ calculate centroid from list of atom indices"""
    coord = []
    for i in atomIndices:
        atom = obmol.GetAtom(i)
        coord.append(atomToCoord(atom))
    return geomutils.averageCoords(coord)


def atomNeighbors(atom):
    """ return atom neighbors"""
    return [x for x in ob.OBAtomAtomIter(a)]


def load_molecule_from_file(fname, molecule_format=None):
    """ load molecule with openbabel"""
    if molecule_format is None:
        n, ftype = utils.getNameExt(fname)
        ftype = ftype.lower()

    mol = ob.OBMol()
    conv = ob.OBConversion()
    conv.SetInFormat(molecule_format)
    conv.ReadFile(mol, fname)

    return mol


def load_molecule_from_string(string, molecule_format):
    """ load molecule with openbabel"""
    mol = ob.OBMol()
    conv = ob.OBConversion()
    conv.SetInFormat(molecule_format)
    conv.ReadString(mol, string)

    return mol


def writeMolecule(mol, fname=None, ftype=None):
    """ save a molecule with openbabel"""
    if ftype is None:
        n, ftype = utils.getNameExt(fname)
        ftype = ftype.lower()

    conv = ob.OBConversion()
    conv.SetOutFormat(ftype)

    if not fname is None:
        conv.WriteFile(mol, fname)
    else:
        return conv.WriteString(mol)


def getPdbInfo(atom):
    """extract information for populating an ATOM/HETATM line
    in the PDB"""
    res = atom.GetResidue()
    if res is None:
        return None
    name = res.GetAtomID(atom)
    chain = res.GetChain()
    resNum = int(res.GetNumString())  # safe way for negative resnumbers
    resName = res.GetName()

    return pdbutils.PDBAtomInfo(name=name, resName=resName, resNum=resNum, chain=chain)


def getPdbInfoNoNull(atom):
    """extract information for populating an ATOM/HETATM line
    in the PDB"""
    res = atom.GetResidue()
    if res is None:
        name = '%-2s' % mini_periodic_table[atom.GetAtomicNum()]
        chain = ' '
        resNum = 1
        resName = 'UNL'
    else:
        name = res.GetAtomID(atom)
        chain = res.GetChain()
        resNum = int(res.GetNumString())  # safe way for negative resnumbers
        resName = res.GetName()
    return pdbutils.PDBAtomInfo(name=name, resName=resName, resNum=resNum, chain=chain)


class OBMolSupplier:
    def __init__(self, fname, _format):
        """  """
        self.fname = fname
        self.conv = ob.OBConversion()
        status = self.conv.SetInFormat(_format)
        if not status:
            raise RuntimeError('could not set OBConversion input format: %s' % _format)
        self.got_mol_in_cache = False
        self.cached_mol = None

    def __iter__(self):
        self.cached_mol = ob.OBMol()
        self.got_mol_in_cache = self.conv.ReadFile(self.cached_mol, self.fname)
        return self

    def __next__(self):
        if self.got_mol_in_cache:
            mol = self.cached_mol
            self.cached_mol = ob.OBMol()
            self.got_mol_in_cache = self.conv.Read(self.cached_mol)
            return mol
        else:
            raise StopIteration





#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#

import argparse
import os
import sys

from rdkit import Chem
from rdkit.six import StringIO

from meeko import PDBQTMolecule


def cmd_lineparser():
    parser = argparse.ArgumentParser(description='Copy atom coordinates from PDBQT (or DLG) file \
                                                  to original molecule file format (SDF or MOL2)')
    parser.add_argument(dest='docking_results_filename',
                        action='store', help='Docking output file to get coordinates. Either a PDBQT \
                        file from Vina or a DLG file from AD-GPU.')
    parser.add_argument('-o', '--output_filename', dest='output_filename',
                        action='store', help='Output molecule filename. If not specified, suffix _docked is \
                        added to the filename based on the input molecule file, and using the same \
                        molecule file format')
    parser.add_argument('-s', '--suffix', dest='suffix_name', default='_docked',
                        action='store', help='Add suffix to output filename if -o/--output_filename \
                        not specified. WARNING: If specified as empty string (\'\'), this will overwrite \
                        the original molecule input file (default: _docked).')
    parser.add_argument('-', '--',  dest='redirect_stdout', action='store_true',
                        help='do not write file, redirect output to STDOUT. Arguments -o/--output_filename \
                        is ignored.')
    return parser.parse_args()


if __name__ == '__main__':
    args = cmd_lineparser()
    docking_results_filename = args.docking_results_filename
    output_filename = args.output_filename
    suffix_name = args.suffix_name
    redirect_stdout = args.redirect_stdout

    output_string = ''

    is_dlg = docking_results_filename.endswith('.dlg')
    pdbqt_mol = PDBQTMolecule.from_file(docking_results_filename, is_dlg=is_dlg, skip_typing=True)

    # RDKit mol from SMILES in docking output PDBQT remarks
    if pdbqt_mol._pose_data['smiles'] is None:
        msg = "\n\n    \"REMARK SMILES\" not found in %s.\n" % docking_results_filename
        msg += "    Consider using -i/--original_input\n"
        raise RuntimeError(msg)
    sio = StringIO()
    f = Chem.SDWriter(sio)
    for pose in pdbqt_mol:
        rdmol = pose.export_rdkit_mol()
        f.write(rdmol)
    f.close()
    output_string += sio.getvalue()
    output_format = 'sdf'

    if not redirect_stdout:
        if output_filename is None:
            output_filename = '%s%s.%s' % (os.path.splitext(docking_results_filename)[0], suffix_name, output_format)

        print(output_string, file=open(output_filename, 'w'))
    else:
        print(output_string)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Meeko PDBQT writer
#

import sys
import json

import numpy as np
from rdkit import Chem
from .utils import pdbutils
from .utils.rdkitutils import mini_periodic_table


class PDBQTWriterLegacy():
    def __init__(self):
        """Initialize the PDBQT writer."""
        self._count = 1
        self._visited = []
        self._numbering = {}
        self._pdbqt_buffer = []
        self._resinfo_set = set() # for flexres keywords BEGIN_RES / END_RES

    def _get_pdbinfo_fitting_pdb_chars(self, pdbinfo):
        """ return strings and integers that are guaranteed
            to fit within the designated chars of the PDB format """

        atom_name = pdbinfo.name
        res_name = pdbinfo.resName
        res_num = pdbinfo.resNum
        chain = pdbinfo.chain
        if len(atom_name) > 4: atom_name = atom_name[0:4]
        if len(res_name) > 3: res_name = res_name[0:3]
        if res_num > 9999: res_num = res_num % 10000
        if len(chain) > 1: chain = chain[0:1]
        return atom_name, res_name, res_num, chain

    def _make_pdbqt_line(self, atom_idx):
        """ """
        record_type = "ATOM"
        alt_id = " "
        pdbinfo = self.setup.pdbinfo[atom_idx]
        if pdbinfo is None:
            pdbinfo = pdbutils.PDBAtomInfo('', '', 0, '')
        resinfo = pdbutils.PDBResInfo(pdbinfo.resName, pdbinfo.resNum, pdbinfo.chain)
        self._resinfo_set.add(resinfo)
        atom_name, res_name, res_num, chain = self._get_pdbinfo_fitting_pdb_chars(pdbinfo)
        in_code = ""
        occupancy = 1.0
        temp_factor = 0.0
        coord = self.setup.coord[atom_idx]
        atom_type = self.setup.get_atom_type(atom_idx)
        charge = self.setup.charge[atom_idx]
        atom = "{:6s}{:5d} {:^4s}{:1s}{:3s} {:1s}{:4d}{:1s}   {:8.3f}{:8.3f}{:8.3f}{:6.2f}{:6.2f}    {:6.3f} {:<2s}"

        return atom.format(record_type, self._count, pdbinfo.name, alt_id, res_name, chain,
                           res_num, in_code, float(coord[0]), float(coord[1]), float(coord[2]),
                           occupancy, temp_factor, charge, atom_type)

    def _walk_graph_recursive(self, node, edge_start=0, first=False): #, rigid_body_id=None):
        """ recursive walk of rigid bodies"""
        if first:
            self._pdbqt_buffer.append('ROOT')
            member_pool = sorted(self.model['rigid_body_members'][node])
        else:
            member_pool = self.model['rigid_body_members'][node][:]
            member_pool.remove(edge_start)
            member_pool = [edge_start] + member_pool

        for member in member_pool:
            if self.setup.atom_ignore[member] == 1:
                continue

            self._pdbqt_buffer.append(self._make_pdbqt_line(member))
            self._numbering[member] = self._count # _count starts at 1
            self._count += 1

        if first:
            self._pdbqt_buffer.append('ENDROOT')

        self._visited.append(node)

        for neigh in self.model['rigid_body_graph'][node]:
            if neigh in self._visited:
                continue

            # Write the branch
            begin, next_index = self.model['rigid_body_connectivity'][node, neigh]

            # do not write branch (or anything downstream) if any of the two atoms
            # defining the rotatable bond are ignored
            if self.setup.atom_ignore[begin] or self.setup.atom_ignore[next_index]:
                continue

            begin = self._numbering[begin]
            end = self._count

            self._pdbqt_buffer.append("BRANCH %3d %3d" % (begin, end))
            self._walk_graph_recursive(neigh, edge_start=next_index)
            self._pdbqt_buffer.append("ENDBRANCH %3d %3d" % (begin, end))

    def write_string(self, setup, add_index_map=False, remove_smiles=False):
        """Output a PDBQT file as a string.

        Args:
            setup: MoleculeSetup

        Returns:
            str: PDBQT string of the molecule

        """
        self._count = 1
        self._visited = []
        self._numbering = {}
        self._pdbqt_buffer = []
        self._atom_counter = {}
        self._resinfo_set = set()

        self.setup = setup
        self.model = setup.flexibility_model
        # get a copy of the current setup, since it's going to be messed up by the hacks for legacy, D3R, etc...
        self.setup = setup.copy()

        root = self.model['root']
        torsdof = len(self.model['rigid_body_graph']) - 1

        if 'torsions_org' in self.model:
            torsdof_org = self.model['torsions_org']
            self._pdbqt_buffer.append('REMARK Flexibility Score: %2.2f' % self.model['score'] )
            active_tors = torsdof_org
        else:
            active_tors = torsdof

        self._walk_graph_recursive(root, first=True)

        if add_index_map:
            for i, remark_line in enumerate(self.remark_index_map()):
                # need to use 'insert' because self._numbering is calculated
                # only after self._walk_graph_recursive
                self._pdbqt_buffer.insert(i, remark_line)

        if not remove_smiles:
            smiles, order = self.setup.get_smiles_and_order()
            missing_h = [] # hydrogens which are not in the smiles
            strings_h_parent = []
            for key in self._numbering:
                if key in self.setup.atom_pseudo: continue
                if key not in order:
                    element = self.setup.get_element(key)
                    #atom = self.setup.mol.GetAtomWithIdx(key)
                    if element != 1:
                        raise RuntimeError("non-Hydrogen atom unexpectedely missing from smiles!?")
                    missing_h.append(key)
                    parents = self.setup.get_neigh(key)
                    parents = [i for i in parents if i < self.setup.atom_true_count] # exclude pseudos
                    if len(parents) != 1:
                        raise RuntimeError("expected hydrogen to have exactly one parent")
                    parent_idx = order[parents[0]] # already 1-indexed
                    string = ' %d %d' % (parent_idx, self._numbering[key]) # key 0-indexed; _numbering[key] 1-indexed
                    strings_h_parent.append(string)
            remarks_h_parent = self.break_long_remark_lines(strings_h_parent, "REMARK H PARENT")
            remark_prefix = "REMARK SMILES IDX"
            remark_idxmap = self.remark_index_map(order, remark_prefix, missing_h)
            remarks = []
            remarks.append("REMARK SMILES %s" % smiles) # break line at 79 chars?
            remarks.extend(remark_idxmap)
            remarks.extend(remarks_h_parent)

            for i, remark_line in enumerate(remarks):
                # need to use 'insert' because self._numbering is calculated
                # only after self._walk_graph_recursive
                self._pdbqt_buffer.insert(i, remark_line)

        if False: #self.setup.is_protein_sidechain:
            if len(self._resinfo_set) > 1:
                print("Warning: more than a single resName, resNum, chain in flexres", file=sys.stderr)
                print(self._resinfo_set, file=sys.stderr)
            resinfo = list(self._resinfo_set)[0]
            pdbinfo = pdbutils.PDBAtomInfo('', resinfo.resName, resinfo.resNum, resinfo.chain)
            _, res_name, res_num, chain = self._get_pdbinfo_fitting_pdb_chars(pdbinfo)
            resinfo_string = "{:3s} {:1s}{:4d}".format(res_name, chain, res_num)
            self._pdbqt_buffer.insert(0, 'BEGIN_RES %s' % resinfo_string)
            self._pdbqt_buffer.append('END_RES %s' % resinfo_string)
        else: # no TORSDOF in flexres
            # torsdof is always going to be the one of the rigid, non-macrocyclic one
            self._pdbqt_buffer.append('TORSDOF %d' % active_tors)


        return '\n'.join(self._pdbqt_buffer) + '\n'


    def remark_index_map(self, order=None, prefix="REMARK INDEX MAP", missing_h=[]):
        """ write mapping of atom indices from input molecule to output PDBQT
            order[ob_index(i.e. 'key')] = smiles_index
        """

        if order is None: order = {key: key+1 for key in self._numbering} # FIXME key+1 breaks OB
        #max_line_length = 79
        #remark_lines = []
        #line = prefix
        strings = []
        for key in self._numbering:
            if key in self.setup.atom_pseudo: continue
            if key in missing_h: continue
            string = " %d %d" % (order[key], self._numbering[key])
            strings.append(string)
        return self.break_long_remark_lines(strings, prefix)
        #    candidate_text = " %d %d" % (order[key], self._numbering[key])
        #    if (len(line) + len(candidate_text)) < max_line_length:
        #        line += candidate_text
        #    else:
        #        remark_lines.append(line)
        #        line = 'REMARK INDEX MAP' + candidate_text
        #remark_lines.append(line)
        #return remark_lines

    def break_long_remark_lines(self, strings, prefix, max_line_length=79):
        remarks = [prefix]
        for string in strings:
            if (len(remarks[-1]) + len(string)) < max_line_length:
                remarks[-1] += string
            else:
                remarks.append(prefix + string)
        return remarks



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Meeko - Interactions
#

import os
from abc import ABC, abstractmethod

import numpy as np


def _compute_angle(v1, v2):
    unit_vector_1 = v1 / np.linalg.norm(v1)
    unit_vector_2 = v2 / np.linalg.norm(v2)
    dot_product = np.dot(unit_vector_1, unit_vector_2)
    angle = np.arccos(dot_product)
    return angle


class _Interaction(ABC):
    """Abstract class for molecular interactions"""
    @abstractmethod
    def find(self, **kwargs):
        pass

    @abstractmethod
    def name(self):
        pass


class _DistanceBased(_Interaction):
    """Generic class for interactions (hydrohpbic, metal, reactive, etc...) based on distance only"""
    def __init__(self, distance, atom_properties):
        assert distance > 0, 'Distance must be superior than 0 Angstrom'
        assert len(atom_properties) == 2, 'An array-like of two atom properties must be defined'

        self._distance = distance
        self._atom_properties = atom_properties

    @property
    def name(self):
        return type(self).__name__

    def find(self, ligand, receptor):
        """Find distance-based interactions between ligand and receptor.

        Args:
            ligand (PDBQTMolecularr): ligand
            receptor (PDBQTReceptor): receptor

        Returns:
            np.ndarray: array of (ligand_atom_index, receptor_atom_index, distance)
                for each interaction between ligand and rigid receptor
            np.ndarray: array of (ligand_atom_index, receptor_atom_index, distance)
                for each interaction between ligand and flexible sidechain

        """
        dtype = [('ligand_idx', int), ('receptor_idx', int), 
                 ('distance', float)]
        rigid_interactions = [] 
        flex_interactions = []
        has_flexible_residues = ligand.has_flexible_residues()
        lig_atoms = ligand.atoms_by_properties(['ligand', self._atom_properties[0]])

        for lig_atom in lig_atoms:
            # Get interactions with the rigid part of the receptor
            rec_atoms = receptor.closest_atoms_from_positions(lig_atom['xyz'], self._distance, self._atom_properties[1])
            rigid_interactions.extend((lig_atom['idx'], rec_atom['idx'], np.linalg.norm(lig_atom['xyz'] - rec_atom['xyz'])) for rec_atom in rec_atoms)

            # Get interactions with the flexible part of the receptor (if present)
            if has_flexible_residues:
                rec_atoms = ligand.closest_atoms_from_positions(lig_atom['xyz'], self._distance, ['flexible_residue', self._atom_properties[1]])
                flex_interactions.extend((lig_atom['idx'], rec_atom['idx'], np.linalg.norm(lig_atom['xyz'] - rec_atom['xyz'])) for rec_atom in rec_atoms)

        return np.array(rigid_interactions, dtype=dtype), np.array(flex_interactions, dtype=dtype)


class Hydrophobic(_DistanceBased):
    def __init__(self, distance=4.5):
        """Hydrophobic interactions

        Args:
            distance (float): distance cutoff between ligand and receptor/flexible sidechain atoms (default: 4.5)

        """
        super().__init__(distance, ['vdw', 'all'])


class Reactive(_DistanceBased):
    def __init__(self, distance=2.0):
        """Reactive interaction between ligand and receptor (flexible sidechain).

        Args:
            distance (float): distance cutoff between ligand and reactive (flexible) sidechain atoms (default: 2.0)

        """
        super().__init__(distance, ['reactive', 'reactive'])


class Metal(_DistanceBased):
    def __init__(self, distance=3.0):
        """Metal interaction between the ligand and metals in the receptor.
        
        Args:
            distance (float): distance cutoff between ligand and metal receptor atoms (default: 3.0)

        """
        super().__init__(distance, ['hb_acc', 'metal'])


class _HBBased(_Interaction):
    """Generic class for hydrogen bond-like interactions"""
    def __init__(self, distance, angles, atom_properties):
        assert distance > 0, 'Distance must be superior than 0 Angstrom'
        assert len(angles) == 2, 'An array-like of two angles must be defined'
        assert len(atom_properties) == 2, 'An array-like of two atom properties must be defined'

        self._distance = distance
        self._angles = [np.radians(angle) for angle in angles]
        self._atom_properties = atom_properties

    @property
    def name(self):
        return type(self).__name__

    def find(self, ligand, receptor):
        """Find Hydrogen bond-like interactions between ligand and receptor.

        Args:
            ligand (PDBQTMolecularr): ligand
            receptor (PDBQTReceptor): receptor

        Returns:
            np.ndarray: array of (ligand_atom_index, receptor_atom_index, distance, donor_angle, acceptor_angle)
                for each interaction between ligand and rigid receptor
            np.ndarray: array of (ligand_atom_index, receptor_atom_index, distance, donor_angle, acceptor_angle)
                for each interaction between ligand and flexible sidechain

        """
        dtype = [('ligand_idx', int), ('receptor_idx', int), 
                 ('distance', float), ('angle_don', float), ('angle_acc', float)]
        rigid_interactions = []
        flex_interactions = []
        lig_hb_pre_position = None
        rec_hb_pre_position = None
        lig_atom_property, rec_atom_property = self._atom_properties

        has_flexible_residues = ligand.has_flexible_residues()

        if lig_atom_property == 'water':
            lig_atoms = ligand.atoms_by_properties([lig_atom_property])
        else:
            lig_atoms = ligand.atoms_by_properties(['ligand', lig_atom_property])

        for lig_atom in lig_atoms:
            # Get pre-acceptor position (if acceptor) or pre-hydrogen position (if donor) for that atom in the ligand
            lig_bound_atoms_index = ligand.neighbor_atoms(lig_atom['idx'])
            if any(lig_bound_atoms_index):
                lig_bound_atoms = ligand.atoms(lig_bound_atoms_index[0])
                # This is not accurate when bonds don't have the same length
                lig_hb_pre_position = np.mean(lig_bound_atoms['xyz'], axis=0)
            else:
                # If no atom bound, likely a water molecule attached to the ligand
                lig_hb_pre_position = None

            # Get rigid part of the receptor
            rec_rigid_atoms = receptor.closest_atoms_from_positions(lig_atom['xyz'], self._distance, rec_atom_property)
            rec_rigid_flex = [receptor]
            rec_rigid_flex_atoms = [rec_rigid_atoms]
            rec_rigid_flex_types = ['rigid']

            # Get the flexible part of the receptor (if present)
            if has_flexible_residues:
                rec_flex_atoms = ligand.closest_atoms_from_positions(lig_atom['xyz'], self._distance, ['flexible_residue', rec_atom_property])
                rec_rigid_flex.append(ligand)
                rec_rigid_flex_atoms.append(rec_flex_atoms)
                rec_rigid_flex_types.append('flex')

            # Add interactions
            for rec, rec_atoms, rec_type in zip(rec_rigid_flex, rec_rigid_flex_atoms, rec_rigid_flex_types):
                for rec_atom in rec_atoms:
                    distance = np.linalg.norm(lig_atom['xyz'] - rec_atom['xyz'])

                    if distance <= self._distance:
                        # Get pre-acceptor position (if acceptor) or pre-hydrogen position (if donor) for that atom in the ligand
                        rec_bound_atoms_index = rec.neighbor_atoms(rec_atom['idx'])
                        if any(rec_bound_atoms_index):
                            rec_bound_atoms = rec.atoms(rec_bound_atoms_index[0])
                            # This is not accurate when bonds don't have the same length
                            rec_hb_pre_position = np.mean(rec_bound_atoms['xyz'], axis=0)
                        else:
                            # If no atom bound, likely a water molecule in the receptor
                            rec_hb_pre_position = None

                        if lig_hb_pre_position is not None:
                            angle_1 = _compute_angle(lig_atom['xyz'] - lig_hb_pre_position, lig_atom['xyz'] - rec_atom['xyz'])
                        else:
                            angle_1 = np.radians(180)

                        if rec_hb_pre_position is not None:
                            angle_2 = _compute_angle(rec_atom['xyz'] - rec_hb_pre_position, rec_atom['xyz'] - lig_atom['xyz'])
                        else:
                            angle_2 = np.radians(180)

                        if (angle_1 >= self._angles[0]) & (angle_2 >= self._angles[1]):
                            tmp = [lig_atom['idx'], rec_atom['idx'], distance]
                            if lig_atom_property == 'hb_don':
                                tmp += [np.degrees(angle_1), np.degrees(angle_2)]
                            else:
                                tmp += [np.degrees(angle_2), np.degrees(angle_1)]

                            if rec_type == 'rigid':
                                rigid_interactions.append(tuple(tmp))
                            else:
                                flex_interactions.append(tuple(tmp))

        return np.array(rigid_interactions, dtype=dtype), np.array(flex_interactions, dtype=dtype)


class HBDonor(_HBBased):
    def __init__(self, distance=3.5, angles=(120, 90)):
        """Hydrogen bond interaction between the ligand (donor) and the receptor (acceptor)
        
        Args:
            distance (float): distance cutoff between donor and acceptor atoms (default: 3.5)
            angles (array-like): angles between donor and acceptor atoms (default: (120, 90))
                Donor-H -- acceptor angle              (angle_1)
                Pre_acceptor-acceptor -- H angle       (angle_2)

        """
        # Because the distance will be between the acceptor and the hydrogen atoms
        super().__init__(distance - 1., angles, ['hb_don', 'hb_acc'])


class HBAcceptor(_HBBased):
    def __init__(self, distance=3.5, angles=(120, 90)):
        """Hydrogen bond interaction between the ligand (acceptor) and the receptor (donor)
        
        Args:
            distance (float): distance cutoff between donor and acceptor atoms (default: 3.5)
            angles (array-like): angles between donor and acceptor atoms (default: (120, 90))
                Donor-H -- acceptor angle              (angle_1)
                Pre_acceptor-acceptor -- H angle       (angle_2)

        """
        # Because the distance will be between the acceptor and the hydrogen atoms
        super().__init__(distance - 1., angles[::-1], ['hb_acc', 'hb_don'])


class WaterDonor(_HBBased):
    def __init__(self, distance=3.2, angle=90):
        """Interaction between donor water molecules (attached to the ligand) and receptor (acceptor)
        
        Args:
            distance (float): distance cutoff between donor water molecule and acceptor atoms (default: 3.2)
            angle (float): angle between donor water molecule and acceptor atoms (default: 90)
                W_donor -- Pre_acceptor-acceptor angle

        """
        super().__init__(distance, (0, angle), ['water', 'hb_acc'])


class WaterAcceptor(_HBBased):
    def __init__(self, distance=3.2, angle=120):
        """Interaction between acceptor water molecules (attached to the ligand) and receptor (donor)
        
        Args:
            distance (float): distance cutoff between donor water molecule and acceptor atoms (default: 3.2)
            angle (float): angle between donor water molecule and acceptor atoms (default: 120)
                W_acceptor -- Donor-H angle

        """
        # Because the distance will be between the acceptor and the hydrogen atoms
        super().__init__(distance - 1., (0, angle), ['water', 'hb_don'])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Meeko
#

try:
    import openbabel
except ImportError:
    _has_openbabel = False
else:
    _has_openbabel = True

try:
    import prody
except ImportError:
    _has_prody = False
else:
    _has_prody = True

from .preparation import MoleculePreparation
from .molsetup import RDKitMoleculeSetup
from .utils import rdkitutils
from .utils import pdbutils
from .utils import geomutils
from .utils import utils
from .atomtyper import AtomTyper
from .receptor_pdbqt import PDBQTReceptor
from .molecule_pdbqt import PDBQTMolecule
from . import analysis

__all__ = ['MoleculePreparation', 'RDKitMoleculeSetup',
        'pdbutils', 'geomutils', 'rdkitutils', 'utils',
        'AtomTyper', 'PDBQTMolecule', 'PDBQTReceptor', 'analysis']

if _has_openbabel:
    from .molsetup import OBMoleculeSetup
    from .utils import obutils
    __all__.append("OBMoleculeSetup")
    __all__.append("obutils")

if _has_prody:
    from .covalentbuilder import CovalentBuilder
    __all__.append("CovalentBuilder")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Meeko preparation
#

import os
import sys
from collections import OrderedDict
import warnings

from rdkit import Chem

from .molsetup import OBMoleculeSetup
from .molsetup import RDKitMoleculeSetup
from .atomtyper import AtomTyper
from .bondtyper import BondTyperLegacy
from .hydrate import HydrateMoleculeLegacy
from .macrocycle import FlexMacrocycle
from .flexibility import FlexibilityBuilder
from .writer import PDBQTWriterLegacy

try:
    from openbabel import openbabel as ob
except ImportError:
    _has_openbabel = False
else:
    _has_openbabel = True


class MoleculePreparation:
    def __init__(self, keep_nonpolar_hydrogens=False,
            hydrate=False, flexible_amides=False,
            rigid_macrocycles=False, min_ring_size=7, max_ring_size=33,
            keep_chorded_rings=False, keep_equivalent_rings=False,
            rigidify_bonds_smarts=[], rigidify_bonds_indices=[],
            double_bond_penalty=50, atom_type_smarts={},
            add_index_map=False,
            stop_at_defaults=False, remove_smiles=False):

        self.keep_nonpolar_hydrogens = keep_nonpolar_hydrogens
        self.hydrate = hydrate
        self.flexible_amides = flexible_amides
        self.rigid_macrocycles = rigid_macrocycles
        self.min_ring_size = min_ring_size
        self.max_ring_size = max_ring_size
        self.keep_chorded_rings = keep_chorded_rings
        self.keep_equivalent_rings = keep_equivalent_rings
        self.rigidify_bonds_smarts = rigidify_bonds_smarts
        self.rigidify_bonds_indices = rigidify_bonds_indices
        self.double_bond_penalty = double_bond_penalty
        self.atom_type_smarts = atom_type_smarts
        self.add_index_map = add_index_map
        self.remove_smiles = remove_smiles

        if stop_at_defaults: return # create an object to show just the defaults (e.g. to argparse)

        self.setup = None
        self._atom_typer = AtomTyper(self.atom_type_smarts)
        self._bond_typer = BondTyperLegacy()
        self._macrocycle_typer = FlexMacrocycle(
                self.min_ring_size, self.max_ring_size, self.double_bond_penalty)
        self._flex_builder = FlexibilityBuilder()
        self._water_builder = HydrateMoleculeLegacy()
        self._writer = PDBQTWriterLegacy()
        self.is_ok = None
        self.log = None
        self._classes_setup = {Chem.rdchem.Mol: RDKitMoleculeSetup}
        if _has_openbabel:
            self._classes_setup[ob.OBMol] = OBMoleculeSetup
        if keep_chorded_rings and keep_equivalent_rings==False:
            warnings.warn("keep_equivalent_rings=False ignored because keep_chorded_rings=True", RuntimeWarning)

    @classmethod
    def init_just_defaults(cls):
        return cls(stop_at_defaults=True)

    @ classmethod
    def from_config(cls, config):
        expected_keys = cls.init_just_defaults().__dict__.keys()
        bad_keys = [k for k in config if k not in expected_keys]
        for key in bad_keys:
            print("ERROR: unexpected key \"%s\" in MoleculePreparation.from_config()" % key, file=sys.stderr)
        if len(bad_keys) > 0:
            raise ValueError
        p = cls(**config)
        return p

    def prepare(self, mol, root_atom_index=None, not_terminal_atoms=[]):
        """ if protein_sidechain, C H N O will be removed,
            root will be CA, and BEGIN/END_RES will be added.
        """
        mol_type = type(mol)
        if not mol_type in self._classes_setup:
            raise TypeError("Molecule is not an instance of supported types: %s" % type(mol))
        setup_class = self._classes_setup[mol_type]
        setup = setup_class(mol,
            keep_chorded_rings=self.keep_chorded_rings,
            keep_equivalent_rings=self.keep_equivalent_rings)
        self.setup = setup
        # 1.  assign atom types (including HB types, vectors and stuff)
        # DISABLED TODO self.atom_typer.set_parm(mol)
        self._atom_typer(setup)
        # 2a. add pi-model + merge_h_pi (THIS CHANGE SOME ATOM TYPES)
        # disabled
        # 2b. merge_h_classic
        if not self.keep_nonpolar_hydrogens:
            setup.merge_hydrogen()
        # 3.  assign bond types by using SMARTS...
        #     - bonds should be typed even in rings (but set as non-rotatable)
        #     - if macrocycle is selected, they will be enabled (so they must be typed already!)
        self._bond_typer(setup, self.flexible_amides, self.rigidify_bonds_smarts, self.rigidify_bonds_indices, not_terminal_atoms)
        # 4 . hydrate molecule
        if self.hydrate:
            self._water_builder.hydrate(setup)
        # 5.  break macrocycles into open/linear form
        if self.rigid_macrocycles:
            break_combo_data = None
            bonds_in_rigid_rings = None # not true, but this is only needed when breaking macrocycles
        else:
            break_combo_data, bonds_in_rigid_rings = self._macrocycle_typer.search_macrocycle(setup)

        # 6.  build flexibility...
        # 6.1 if macrocycles typed:
        #     - walk the setup graph by skipping proposed closures
        #       and score resulting flex_trees basing on the lenght
        #       of the branches generated
        #     - actually break the best closure bond (THIS CHANGES SOME ATOM TYPES)
        # 6.2  - walk the graph and build the flextree
        # 7.  but disable all bonds that are in rings and not
        #     in flexible macrocycles
        # TODO restore legacy AD types for PDBQT
        #self._atom_typer.set_param_legacy(mol)

        new_setup = self._flex_builder(setup,
                                       root_atom_index=root_atom_index,
                                       break_combo_data=break_combo_data,
                                       bonds_in_rigid_rings=bonds_in_rigid_rings)

        self.setup = new_setup
        # TODO re-run typing after breaking bonds
        # self.bond_typer.set_types_legacy(mol, exclude=[macrocycle_bonds])
        self.is_ok = self._check()


    def _check(self):
        # verify that all atoms have been typed
        is_ok = True
        msg = ""
        for idx in self.setup.atom_type:
            atom_type = self.setup.atom_type[idx]
            if atom_type is None:
                msg += 'atom number %d has None type, mol name: %s' % (idx, self.setup.get_mol_name())
                is_ok = False
        self.log = msg
        return is_ok 


    def show_setup(self):
        if self.setup is not None:
            tot_charge = 0

            print("Molecule setup\n")
            print("==============[ ATOMS ]===================================================")
            print("idx  |          coords            | charge |ign| atype    | connections")
            print("-----+----------------------------+--------+---+----------+--------------- . . . ")
            for k, v in list(self.setup.coord.items()):
                print("% 4d | % 8.3f % 8.3f % 8.3f | % 1.3f | %d" % (k, v[0], v[1], v[2],
                      self.setup.charge[k], self.setup.atom_ignore[k]),
                      "| % -8s |" % self.setup.atom_type[k],
                      self.setup.graph[k])
                tot_charge += self.setup.charge[k]
            print("-----+----------------------------+--------+---+----------+--------------- . . . ")
            print("  TOT CHARGE: %3.3f" % tot_charge)

            print("\n======[ DIRECTIONAL VECTORS ]==========")
            for k, v in list(self.setup.coord.items()):
                if k in self.setup.interaction_vector:
                    print("% 4d " % k, self.setup.atom_type[k], end=' ')

            print("\n==============[ BONDS ]================")
            # For sanity users, we won't show those keys for now
            keys_to_not_show = ['bond_order', 'type']
            for k, v in list(self.setup.bond.items()):
                t = ', '.join('%s: %s' % (i, j) for i, j in v.items() if not i in keys_to_not_show)
                print("% 8s - " % str(k), t)

            self._macrocycle_typer.show_macrocycle_scores(self.setup)

            print('')

    def write_pdbqt_string(self, add_index_map=None, remove_smiles=None):
        if self.is_ok == False:
            raise RuntimeError("Molecule not OK, refusing to write PDBQT\n\nLOG:\n%s" % self.log)
        if add_index_map is None: add_index_map = self.add_index_map
        if remove_smiles is None: remove_smiles = self.remove_smiles
        if self.setup is not None:
            return self._writer.write_string(self.setup, add_index_map, remove_smiles)
        else:
            raise RuntimeError('Cannot generate PDBQT file, the molecule is not prepared.')

    def write_pdbqt_file(self, pdbqt_filename, add_index_map=None, remove_smiles=None):
        with open(pdbqt_filename,'w') as w:
            w.write(self.write_pdbqt_string(add_index_map, remove_smiles))

    def adapt_pdbqt_for_autodock4_flexres(self, pdbqt_string, res, chain, num):
        """ adapt pdbqt_string to be compatible with AutoDock4 requirements:
             - first and second atoms named CA and CB
             - write BEGIN_RES / END_RES
             - remove TORSDOF
            this is for covalent docking (tethered)
        """
        new_string = "BEGIN_RES %s %s %s\n" % (res, chain, num)
        atom_number = 0
        for line in pdbqt_string.split("\n"):
            if line == "":
                continue
            if line.startswith("TORSDOF"):
                continue
            if line.startswith("ATOM"):
                atom_number+=1
                if atom_number == 1:
                    line = line[:13] + 'CA' + line[15:]
                elif atom_number == 2:
                    line = line[:13] + 'CB' + line[15:]
                new_string += line + '\n'
                continue
            new_string += line + '\n'
        new_string += "END_RES %s %s %s\n" % (res, chain, num)
        return new_string




#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Meeko
#

import os
from collections import defaultdict

import numpy as np
from scipy import spatial

from .utils.covalent_radius_table import covalent_radius
from .utils.autodock4_atom_types_elements import autodock4_atom_types_elements


atom_property_definitions = {'H': 'vdw', 'C': 'vdw', 'A': 'vdw', 'N': 'vdw', 'P': 'vdw', 'S': 'vdw',
                             'Br': 'vdw', 'I': 'vdw', 'F': 'vdw', 'Cl': 'vdw',
                             'NA': 'hb_acc', 'OA': 'hb_acc', 'SA': 'hb_acc', 'OS': 'hb_acc', 'NS': 'hb_acc',
                             'HD': 'hb_don', 'HS': 'hb_don',
                             'Mg': 'metal', 'Ca': 'metal', 'Fe': 'metal', 'Zn': 'metal', 'Mn': 'metal',
                             'MG': 'metal', 'CA': 'metal', 'FE': 'metal', 'ZN': 'metal', 'MN': 'metal',
                             'W': 'water',
                             'G0': 'glue', 'G1': 'glue', 'G2': 'glue', 'G3': 'glue',
                             'CG0': 'glue', 'CG1': 'glue', 'CG2': 'glue', 'CG3': 'glue'}


def _read_receptor_pdbqt_file(pdbqt_filename):
    i = 0
    atoms = []
    atoms_dtype = [('idx', 'i4'), ('serial', 'i4'), ('name', 'U4'), ('resid', 'i4'),
                   ('resname', 'U3'), ('chain', 'U1'), ("xyz", "f4", (3)),
                   ('partial_charges', 'f4'), ('atom_type', 'U2')]
    atom_annotations = {'hb_acc': [], 'hb_don': [],
                        'all': [], 'vdw': [],
                        'metal': []}
    # TZ is a pseudo atom for AutoDock4Zn FF
    pseudo_atom_types = ['TZ']

    with open(pdbqt_filename) as f:
        lines = f.readlines()

        for line in lines:
            if line.startswith('ATOM') or line.startswith("HETATM"):
                idx = i
                serial = int(line[6:11].strip())
                name = line[12:16].strip()
                resname = line[17:20].strip()
                chainid = line[21].strip()
                resid = int(line[22:26].strip())
                xyz = np.array([line[30:38].strip(), line[38:46].strip(), line[46:54].strip()], dtype=np.float32)
                partial_charges = float(line[71:77].strip())
                atom_type = line[77:79].strip()

                if not atom_type in pseudo_atom_types:
                    atom_annotations['all'].append(i)
                    atom_annotations[atom_property_definitions[atom_type]].append(i)
                    atoms.append((idx, serial, name, resid, resname, chainid, xyz, partial_charges, atom_type))

                i += 1

    atoms = np.array(atoms, dtype=atoms_dtype)

    return atoms, atom_annotations


def _identify_bonds(atom_idx, positions, atom_types):
    bonds = defaultdict(list)
    KDTree = spatial.cKDTree(positions)
    bond_allowance_factor = 1.1
    # If we ask more than the number of coordinates/element
    # in the BHTree, we will end up with some inf values
    k = 5 if len(atom_idx) > 5 else len(atom_idx)
    atom_idx = np.array(atom_idx)

    for atom_i, position, atom_type in zip(atom_idx, positions, atom_types):
        distances, indices = KDTree.query(position, k=k)
        r_cov = covalent_radius[autodock4_atom_types_elements[atom_type]]

        optimal_distances = [bond_allowance_factor * (r_cov + covalent_radius[autodock4_atom_types_elements[atom_types[i]]]) for i in indices[1:]]
        bonds[atom_i] = atom_idx[indices[1:][np.where(distances[1:] < optimal_distances)]].tolist()

    return bonds


class PDBQTReceptor:

    def __init__(self, pdbqt_filename):
        self._pdbqt_filename = pdbqt_filename
        self._atoms = None
        self._atom_annotations = None
        self._KDTree = None

        self._atoms, self._atom_annotations = _read_receptor_pdbqt_file(self._pdbqt_filename)
        # We add to the KDTree only the rigid part of the receptor
        self._KDTree = spatial.cKDTree(self._atoms['xyz'])
        self._bonds = _identify_bonds(self._atom_annotations['all'], self._atoms['xyz'], self._atoms['atom_type'])

    def __repr__(self):
        return ('<Receptor from PDBQT file %s containing %d atoms>' % (self._pdbqt_filename, self._atoms.shape[0]))

    def atoms(self, atom_idx=None):
        """Return the atom i

        Args:
            atom_idx (int, list): index of one or multiple atoms

        Returns:
            ndarray: 2d ndarray (atom_id, atom_name, resname, resid, chainid, xyz, q, t)

        """
        if atom_idx is not None and self._atoms.size > 1:
            if not isinstance(atom_idx, (list, tuple, np.ndarray)):
                atom_idx = np.array(atom_idx, dtype=np.int)
            atoms = self._atoms[atom_idx]
        else:
            atoms = self._atoms

        return atoms.copy()

    def positions(self, atom_idx=None):
        """Return coordinates (xyz) of all atoms or a certain atom

        Args:
            atom_idx (int, list): index of one or multiple atoms (0-based)

        Returns:
            ndarray: 2d ndarray of coordinates (xyz)

        """
        return np.atleast_2d(self.atoms(atom_idx)['xyz'])

    def closest_atoms_from_positions(self, xyz, radius, atom_properties=None, ignore=None):
        """Retrieve indices of the closest atoms around a positions/coordinates 
        at a certain radius.

        Args:
            xyz (np.ndarray): array of 3D coordinates
            raidus (float): radius
            atom_properties (str): property of the atoms to retrieve 
                (properties: ligand, flexible_residue, vdw, hb_don, hb_acc, metal, water, reactive, glue)
            ignore (int or list): ignore atom for the search using atom id (0-based)

        Returns:
            ndarray: 2d ndarray (atom_id, atom_name, resname, resid, chainid, xyz, q, t)

        """
        index = self._KDTree.query_ball_point(xyz, radius, p=2, return_sorted=True)

        # When nothing was found around...
        if not index:
            return np.array([])

        # Handle the case when positions for of only one atom was passed in the input
        try:
            index = {i for j in index for i in j}
        except:
            index = set(index)

        if atom_properties is not None:
            if not isinstance(atom_properties, (list, tuple)):
                atom_properties = [atom_properties]

            try:
                for atom_property in atom_properties:
                    index.intersection_update(self._atom_annotations[atom_property])
            except:
                error_msg = 'Atom property %s is not valid. Valid atom properties are: %s'
                raise KeyError(error_msg % (atom_property, self._atom_annotations.keys()))

        if ignore is not None:
            if not isinstance(ignore, (list, tuple, np.ndarray)):
                ignore = [ignore]
            index = index.difference([i for i in ignore])

        index = list(index)
        atoms = self._atoms[index].copy()

        return atoms

    def closest_atoms(self, atom_idx, radius, atom_properties=None):
        """Retrieve indices of the closest atoms around a positions/coordinates 
        at a certain radius.

        Args:
            atom_idx (int, list): index of one or multiple atoms (0-based)
            raidus (float): radius
            atom_properties (str or list): property of the atoms to retrieve 
                (properties: ligand, flexible_residue, vdw, hb_don, hb_acc, metal, water, reactive, glue)

        Returns:
            ndarray: 2d ndarray (atom_id, atom_name, resname, resid, chainid, xyz, q, t)

        """
        return self.closest_atoms_from_positions(self._atoms[atom_idx]['xyz'], radius, atom_properties, atom_idx)

    def neighbor_atoms(self, atom_idx):
        """Return neighbor (bonded) atoms

        Args:
            atom_idx (int, list): index of one or multiple atoms (0-based)

        Returns:
            list_of_list: list of lists containing the neighbor (bonded) atoms (0-based)

        """
        if not isinstance(atom_idx, (list, tuple, np.ndarray)):
            atom_idx = [atom_idx]

        return [self._bonds[i] for i in atom_idx]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import fnmatch
from setuptools import setup, find_packages


# Path to the directory that contains this setup.py file.
base_dir = os.path.abspath(os.path.dirname(__file__))

def find_files(directory):
    matches = []

    for root, dirnames, filenames in os.walk(directory):
        for filename in fnmatch.filter(filenames, '*'):
            matches.append(os.path.join(root, filename))

    return matches


setup(
    name="meeko",
    version='0.3.3',
    author="Forli Lab",
    author_email="forli@scripps.edu",
    url="https://github.com/ccsb-scripps/meeko",
    description='Python package for preparing small molecule for docking',
    long_description=open(os.path.join(base_dir, 'README.md')).read(),
    long_description_content_type="text/markdown",
    packages=find_packages(),
    scripts=["scripts/mk_prepare_ligand.py",
             "scripts/mk_copy_coords.py"],
           #"scripts/dry.py",
           #"scripts/mapwater.py",
           #"scripts/wet.py"],
    package_data={"meeko" : ["data/*"]},
    data_files=[("", ["README.md", "LICENSE"]),
                ("scripts", find_files("scripts"))],
    include_package_data=True,
    zip_safe=False,
    install_requires=['numpy>=1.18'],
    python_requires='>=3.5.*',
    license="Apache-2.0",
    keywords=["molecular modeling", "drug design",
            "docking", "autodock"],
    classifiers=[
        'Environment :: Console',
        'Environment :: Other Environment',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Operating System :: MacOS :: MacOS X',
        #'Operating System :: Microsoft :: Windows',
        'Operating System :: OS Independent',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: C++',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Topic :: Scientific/Engineering :: Chemistry',
        'Topic :: Software Development :: Libraries'
    ]
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Meeko hydrate molecule
#

import numpy as np

from .utils import geomutils
from .utils import pdbutils


class HydrateMoleculeLegacy:
    def __init__(self, distance=3.0, charge=0, atom_type="W"):
        """Initialize the legacy hydrate typer for AutoDock 4.2.x

        Args:
            distance (float): distance between water molecules and ligand heavy atoms. (default: 3.0)
            charge (float): partial charge of the water molecule. Not use for the hydrated docking. (default: 0)
            atom_type (str): atom type of the water molecule. (default: W)

        """
        self._distance = distance
        self._charge = charge
        self._atom_type = atom_type
        self._bond_type = 1
        self._rotatable = False
        self._hb_config = {'HD': {1: (1, 1)},    # neigh: 1, wat: 1, sp1
                           'OA': {
                                      1: (2, 2), # neigh: 1, wat: 2, sp2
                                      2: (2, 3)  # neigh: 2, wat: 2, sp3
                                  },
                           'SA': {
                                      1: (2, 2), # neigh: 1, wat: 2, sp2
                                      2: (2, 3)  # neigh: 2, wat: 2, sp3
                                  },
                           'NA': {
                                      1: (1, 1), # neigh: 1, wat: 3, sp1
                                      2: (1, 2), # neigh: 2, wat: 1, sp2
                                      3: (1, 3)  # neigh: 3, wat: 1, sp3
                                  }
                           }

    def _place_sp1_one_water(self, anchor_xyz, neighbor_xyz, hb_length=3.0):
        position = anchor_xyz + geomutils.vector(neighbor_xyz, anchor_xyz)
        position = geomutils.resize_vector(position, hb_length, anchor_xyz)
        positions = np.array([position])

        return positions

    def _place_sp2_one_water(self, anchor_xyz, neighbor1_xyz, neighbor2_xyz, hb_length=3.0):
        position = geomutils.atom_to_move(anchor_xyz, [neighbor1_xyz, neighbor2_xyz])
        position = geomutils.resize_vector(position, hb_length, anchor_xyz)
        positions = np.array([position])

        return positions

    def _place_sp2_two_waters(self, anchor_xyz, neighbor1_xyz, neighbor2_xyz, hb_lengths, angles):
        if len(hb_lengths) != 2:
            raise ValueError()
        if len(angles) != 2:
            raise ValueError()

        positions = []

        r = geomutils.rotation_axis(neighbor1_xyz, anchor_xyz, neighbor2_xyz, origin=anchor_xyz)
        p = neighbor1_xyz

        # We rotate p to get each vectors if necessary
        for hb_length, angle in zip(hb_lengths, angles):
            vector = p
            if angle != 0.:
                position = geomutils.rotate_point(vector, anchor_xyz, r, angle)
            position = geomutils.resize_vector(position, hb_length, anchor_xyz)
            positions.append(position)

        positions = np.array(positions)

        return positions

    def _place_sp3_one_water(self, anchor_xyz, neighbor1_xyz, neighbor2_xyz, neighbor3_xyz, hb_length):
        # We have to normalize bonds, otherwise the water molecule is not well placed
        v1 = anchor_xyz + geomutils.normalize(geomutils.vector(anchor_xyz, neighbor1_xyz))
        v2 = anchor_xyz + geomutils.normalize(geomutils.vector(anchor_xyz, neighbor2_xyz))
        v3 = anchor_xyz + geomutils.normalize(geomutils.vector(anchor_xyz, neighbor3_xyz))

        position = geomutils.atom_to_move(anchor_xyz, [v1, v2, v3])
        position = geomutils.resize_vector(position, hb_length, anchor_xyz)
        positions = np.array([position])

        return positions

    def _place_sp3_two_waters(self, anchor_xyz, neighbor1_xyz, neighbor2_xyz, hb_lengths, angles):
        if len(hb_lengths) != 2:
            raise ValueError()
        if len(angles) != 2:
            raise ValueError()

        positions = []

        v1 = anchor_xyz + geomutils.normalize(geomutils.vector(anchor_xyz, neighbor1_xyz))
        v2 = anchor_xyz + geomutils.normalize(geomutils.vector(anchor_xyz, neighbor2_xyz))

        r = anchor_xyz + geomutils.normalize(geomutils.vector(v1, v2))
        p = geomutils.atom_to_move(anchor_xyz, [v1, v2])

        # We rotate p to get each vectors if necessary
        for hb_length, angle in zip(hb_lengths, angles):
            vector = p
            if angle != 0.:
                position = geomutils.rotate_point(vector, anchor_xyz, r, angle)
            position = geomutils.resize_vector(position, hb_length, anchor_xyz)
            positions.append(position)

        positions = np.array(positions)

        return positions

    def hydrate(self, setup):
        """Add water molecules to the ligand

        Args:
            setup: MoleculeSetup object

        """
        water_anchors = []
        water_positions = []
        # It will be the same distance for all of the water molecules
        hb_length = self._distance

        for a, neighbors in setup.graph.items():
            atom_type = setup.get_atom_type(a)
            anchor_xyz = setup.get_coord(a)
            neighbor1_xyz = setup.get_coord(neighbors[0])
            positions = np.array([])
            n_wat = None
            hyb = None

            if atom_type in self._hb_config:
                try:
                    n_wat, hyb = self._hb_config[atom_type][len(neighbors)]
                except KeyError:
                    raise RuntimeError('Cannot place water molecules on atom %d of type %s with %d neighbors.' % (a, atom_type, len(neighbors)))

                water_anchors.append(a)

            if hyb == 1:
                if n_wat == 1:
                    # Example: X-HD
                    positions = self._place_sp1_one_water(anchor_xyz,
                                                          neighbor1_xyz,
                                                          hb_length - 1.0)
            elif hyb == 2:
                if n_wat == 1:
                    # Example: X-Nitrogen-X
                    neighbor2_xyz = setup.get_coord(neighbors[1])
                    positions = self._place_sp2_one_water(anchor_xyz,
                                                          neighbor1_xyz, neighbor2_xyz,
                                                          hb_length)
                elif n_wat == 2:
                    # Example: C=0 (backbone oxygen)
                    tmp_neighbors = [x for x in setup.get_neigh(neighbors[0]) if not x == a]
                    neighbor2_xyz =  setup.get_coord(tmp_neighbors[0])
                    positions = self._place_sp2_two_waters(anchor_xyz,
                                                           neighbor1_xyz, neighbor2_xyz,
                                                           [hb_length, hb_length],
                                                           [-np.radians(120), np.radians(120)])
                elif n_wat == 3:
                    hyb = 3
            elif hyb == 3:
                if n_wat == 1:
                    # Example: Ammonia
                    neighbor2_xyz = setup.get_coord(neighbors[1])
                    neighbor3_xyz = setup.get_coord(neighbors[2])
                    positions = self._place_sp3_one_water(anchor_xyz,
                                                          neighbor1_xyz, neighbor2_xyz, neighbor3_xyz,
                                                          hb_length)
                elif n_wat == 2:
                    # Example: O-HD (Oxygen in hydroxyl group)
                    neighbor2_xyz = setup.get_coord(neighbors[1])
                    positions = self._place_sp3_two_waters(anchor_xyz,
                                                           neighbor1_xyz, neighbor2_xyz,
                                                           [hb_length, hb_length],
                                                           [-np.radians(60), np.radians(60)])
                elif n_wat == 3:
                    positions = np.array([])

            if positions.size:
                water_positions.append(positions)

        for water_anchor, waters_on_anchor in zip(water_anchors, water_positions):
            for water_on_anchor in waters_on_anchor:
                tmp = setup.pdbinfo[water_anchor]
                pdbinfo = pdbutils.PDBAtomInfo('WAT', tmp.resName, tmp.resNum, tmp.chain)
                setup.add_pseudo(water_on_anchor, self._charge, [water_anchor], self._atom_type,
                                     self._bond_type, self._rotatable, pdbinfo)
